import atexit
import json
import os
import signal
import subprocess
import time
import urllib.request
from contextlib import asynccontextmanager
from enum import Enum
from pathlib import Path

import uvicorn
from connexion import AsyncApp, ConnexionMiddleware
from connexion.options import SwaggerUIOptions
from fastapi import FastAPI, Request
from filelock import SoftFileLock, Timeout
from prefect import flow
from prefect.server.api.server import SubprocessASGIServer
from prefect.settings import (
    PREFECT_DEFAULT_WORK_POOL_NAME,
    PREFECT_SERVER_EPHEMERAL_STARTUP_TIMEOUT_SECONDS,
)
from prefect.utilities.processutils import get_sys_executable
from prefect.workers.process import ProcessWorker
from tqdm import tqdm

from timestep.api.openai.v1.controllers.completions_controller import create_completion
from timestep.config import Settings
from timestep.database import create_db_and_tables

settings = Settings()

app_dir = settings.app_dir
connexion_app = AsyncApp(import_name=__name__)
default_hf_repo_id = settings.default_hf_repo_id
default_llamafile_filename = settings.default_llamafile_filename
default_llamafile_host = settings.default_llamafile_host
default_llamafile_port = settings.default_llamafile_port
default_llamafile_url = f"https://huggingface.co/{default_hf_repo_id}/resolve/main/{default_llamafile_filename}?download=true"
local_llamafile_path = (
    f"{app_dir}/models/{default_hf_repo_id}/{default_llamafile_filename}"
)


@asynccontextmanager
async def lifespan(app: FastAPI):
    os.makedirs(f"{app_dir}/data", exist_ok=True)
    os.makedirs(f"{app_dir}/models", exist_ok=True)
    os.makedirs(f"{app_dir}/work", exist_ok=True)

    fn = Path(f"{app_dir}/data/data.json")

    soft_lock = SoftFileLock(
        f"{str(fn)}.lock",
        is_singleton=True,
        thread_local=False,
        timeout=120,
    )

    subprocess_manager = SubprocessManager()

    worker_count = int(os.getenv("PYTEST_XDIST_WORKER_COUNT", 1))
    worker_name = os.getenv("PYTEST_XDIST_WORKER", "master")
    worker_pid = os.getpid()

    with soft_lock.acquire():
        create_db_and_tables()

        if fn.is_file():
            data = json.loads(fn.read_text())

            data["timestep"][worker_name] = {
                "pid": worker_pid,
                "status": "running",
            }

            fn.write_text(json.dumps(data))

        else:
            # create_db_and_tables()

            if os.path.basename(
                local_llamafile_path
            ) == default_llamafile_filename and not os.path.exists(
                local_llamafile_path
            ):
                os.makedirs(os.path.dirname(local_llamafile_path), exist_ok=True)
                download_with_progress_bar(default_llamafile_url, local_llamafile_path)

            assert os.path.exists(local_llamafile_path)

            # if local_llamafile_path is not executable, make it executable
            if not os.access(local_llamafile_path, os.X_OK):
                os.chmod(local_llamafile_path, 0o755)

            llamafile_pid = subprocess_manager.start(
                args=[
                    "sh",
                    local_llamafile_path,
                    "--host",
                    f"{default_llamafile_host}",
                    "--nobrowser",
                    "--path",
                    "/zip/llama.cpp/server/public",
                    "--port",
                    f"{default_llamafile_port}",
                    "--temp",
                    "0.0",
                ],
                name="llamafiles.default",
            )

            prefect_server_pid = subprocess_manager.start(
                args=[
                    "sh",
                    "-c",
                    "prefect server start",
                ],
                name="prefect.server",
            )

            # prefect_server = SubprocessASGIServer(port=4200)
            # prefect_server.start(timeout=30)
            # prefect_server_pid = prefect_server.server_process.pid

            # print(
            #     f"Waiting {PREFECT_SERVER_EPHEMERAL_STARTUP_TIMEOUT_SECONDS.value()} seconds for Prefect server to start..."
            # )
            # time.sleep(PREFECT_SERVER_EPHEMERAL_STARTUP_TIMEOUT_SECONDS.value())
            print(f"Sleeping for 30 seconds...")
            time.sleep(30)

            prefect_worker_pid = subprocess_manager.start(
                args=[
                    "sh",
                    "-c",
                    "prefect worker start --pool 'default' --type 'process'",
                ],
                name="prefect.worker.default",
            )

            print(f"Sleeping for 30 seconds...")
            time.sleep(30)

            data = {
                "llamafiles": {
                    "default": {
                        "pid": llamafile_pid,
                        "status": "running",
                    }
                },
                "prefect": {
                    "server": {
                        "pid": prefect_server_pid,
                        "status": "running",
                    },
                    "workers": {
                        "default": {
                            "pid": prefect_worker_pid,
                            "status": "running",
                        }
                    },
                },
                "timestep": {
                    worker_name: {
                        "pid": worker_pid,
                        "status": "running",
                    }
                },
            }

            fn.write_text(json.dumps(data))

    # TODO: Filelock doesnt support async, try https://py-filelock.readthedocs.io/en/latest/_modules/filelock/asyncio.html
    agent_flow = await flow.from_source(
        source=str(Path(__file__).parent),
        entrypoint="worker.py:agent_flow",
    )

    await agent_flow.deploy(
        name="agent-flow-deployment",
        work_pool_name="default",
    )

    yield

    with soft_lock.acquire():
        data = json.loads(fn.read_text())

        print(f"Stopping {worker_name} worker process with PID: {worker_pid}")
        data["timestep"][worker_name]["status"] = "stopped"

        stopped_workers = [
            worker
            for worker in data["timestep"].values()
            if worker["status"] == "stopped"
        ]

        if len(stopped_workers) >= worker_count:
            for prefect_worker_id, prefect_worker in data["prefect"]["workers"].items():
                print(
                    f"Stopping Prefect worker {prefect_worker_id} process with PID: {prefect_worker['pid']}"
                )
                subprocess_manager.stop(prefect_worker["pid"])

            print(
                f"Stopping Prefect server process with PID: {data['prefect']['server']['pid']}"
            )
            subprocess_manager.stop(data["prefect"]["server"]["pid"])

            for llamafile_id, llamafile in data["llamafiles"].items():
                print(
                    f"Stopping llamafile {llamafile_id} process with PID: {llamafile['pid']}"
                )
                subprocess_manager.stop(llamafile["pid"])

            fn.unlink()

        else:
            fn.write_text(json.dumps(data))


fastapi_app = FastAPI(
    lifespan=lifespan,
    # servers= # TODO: add server for testing?
)

connexion_app.add_api(
    "api/ap/v1/openapi/openapi.yaml",
    pythonic_params=True,
    resolver_error=501,
    swagger_ui_options=SwaggerUIOptions(serve_spec=False),
)

connexion_app.add_api(
    "api/openai/v1/openapi/openapi.yaml",
    base_path="/openai/v1",
    pythonic_params=True,
    resolver_error=500,
)

fastapi_app.mount("/api", ConnexionMiddleware(app=connexion_app, import_name=__name__))


class Engine(str, Enum):
    copilot_codex = "copilot-codex"


@fastapi_app.get("/v1/engines")
async def list_engines():
    # return [engine.value for engine in Engine]
    return {
        "data": [
            {
                "id": Engine.copilot_codex,
                "name": "Copilot Codex",
                "description": "OpenAI's Codex model, formerly known as GitHub Copilot",
            }
        ]
    }


@fastapi_app.post("/v1/engines/{engine}/completions")
async def create_code_completion(engine: Engine, request: Request):
    print(
        f"=== {__name__}.create_code_completion(engine: Engine, request: Request) ==="
    )
    print(f"engine: {engine}")
    print(f"request: {request}")

    if engine == Engine.copilot_codex:
        try:
            body = await request.json()

        except json.decoder.JSONDecodeError as e:
            print(f"Error: {e}")
            body = {
                "model": "LLaMA_CPP",
                "prompt": "int main() {",
            }

        # body["model"] = "LLaMA_CPP"

        return await create_completion(body, token_info={}, user=None)

    else:
        raise NotImplementedError("Not implemented")

    # @fastapi_app.get("/v2/models/{model_id}")
    # async def create_code_completion(model_id: str, request: Request):
    #     # model_info = models_service.retrieve_model(model_id, token_info, user)
    #     model_info = model_service.retrieve_model(model_id)

    #     print("model_info: ", model_info)

    #     return model_info


def download_with_progress_bar(
    url, filename
):  # TODO: just use from huggingface_hub import hf_hub_download and hf_hub_upload
    # Extract the basename of the file
    base_filename = os.path.basename(filename)

    # Define the reporthook function
    def reporthook(block_num, block_size, total_size):
        if pbar.total is None:
            pbar.total = total_size

        pbar.update(block_size)

    # Create a tqdm progress bar instance with a custom bar_format
    pbar = tqdm(
        unit="B",
        unit_scale=True,
        miniters=1,
        desc=base_filename,
        bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt} [{percentage:.0f}%]",
    )

    # Use urlretrieve to download the file with the reporthook
    urllib.request.urlretrieve(url, filename, reporthook)

    # Close the progress bar
    pbar.close()


class SubprocessManager:
    def start(self, args, name):
        try:
            stdout_log_path = f"stdout.{name}.log"
            stderr_log_path = f"stderr.{name}.log"

            with open(stderr_log_path, "w") as stderr, open(
                stdout_log_path, "w"
            ) as stdout:
                self.process = subprocess.Popen(
                    args,
                    preexec_fn=os.setpgrp,  # Start the process in a new process group
                    stderr=stderr,
                    stdout=stdout,
                )

            print(f"Started {name} process with PID: {self.process.pid}")

            return self.process.pid

        except Exception as e:
            print(f"An error occurred: {e}")

    def stop(self, pid=None):
        if pid:
            os.killpg(os.getpgid(pid), signal.SIGTERM)

        else:
            print("No process to terminate")


def main(*args, **kwargs):
    print(f"=== {__name__}.main(*args, **kwargs) ===")
    print(f"args: {args}")
    print(f"kwargs: {kwargs}")

    cwd = os.getcwd()
    print(f"cwd: {cwd}")

    uvicorn.run(
        f"{__name__}:fastapi_app",
        # host="0.0.0.0",
        host=kwargs.get("host", "0.0.0.0"),
        log_level="info",
        loop="asyncio",
        # port=8000,
        port=kwargs.get("port", 8000),
        # reload=True,
        reload=kwargs.get("dev", False),
        reload_dirs=[
            f"{os.getcwd()}/timestep",
        ],
        workers=1,
    )


if __name__ == "__main__":
    main()



import os
import uuid
from abc import ABC, abstractmethod
from typing import List

from pydantic import ConfigDict
from sqlalchemy import JSON, Column, DateTime, func, text
from sqlmodel import Field, SQLModel, create_engine

from timestep.config import Settings

settings = Settings()
app_dir = settings.app_dir
os.makedirs(f"{app_dir}/data", exist_ok=True)
engine = create_engine(
    f"sqlite:///{app_dir}/data/database.db", echo=True, echo_pool=True
)


class SQLModelMixin(object):
    ## TODO: Investigate inheritance and use cases of TableArtifact from Prefect
    __table_args__ = {"extend_existing": True}

    # created_at = Column(DateTime, server_default=text('NOW()'))
    created_at = Column(DateTime, default=func.now())
    updated_at = Column(DateTime, default=func.now(), onupdate=func.now())


class AgentSQLModel(SQLModel, SQLModelMixin, table=True):
    __tablename__: str = "agents"
    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)
    description: str | None = None
    instructions: str | None = None
    model: str = Field()
    # models: List[str] = Field(sa_column=Column(JSON), default=[])
    # model_config = ConfigDict(protected_namespaces=())
    # model_id: uuid.UUID = Field(foreign_key="models.id")
    name: str | None = None
    tools: List[dict] = Field(sa_column=Column(JSON), default=[])


class MessageSQLModel(SQLModel, SQLModelMixin, table=True):
    __tablename__: str = "messages"
    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)
    attachments: List[dict] = Field(sa_column=Column(JSON), default=[])
    content: List[dict] = Field(sa_column=Column(JSON), default=[])
    role: str = Field()
    status: str = Field()  # status: Literal['in_progress', 'incomplete', 'completed']
    thread_id: uuid.UUID = Field(foreign_key="threads.id")


# class ModelSQLModel(SQLModel, SQLModelMixin, table=True):
#     __tablename__: str = "models"
#     id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)


# class ModelAliasSQLModel(SQLModel, SQLModelMixin, table=True):
#     __tablename__: str = "model_aliases"
#     alias: str = Field(default=None, primary_key=True)
#     model_config = ConfigDict(protected_namespaces=())
#     model_id: uuid.UUID = Field(foreign_key="models.id", ondelete="CASCADE")


class ThreadSQLModel(SQLModel, SQLModelMixin, table=True):
    __tablename__: str = "threads"
    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)


def create_db_and_tables():
    # try:
    SQLModel.metadata.create_all(
        bind=engine,
        checkfirst=True,
        tables=None,
    )

    # except Exception as e:
    #     print(f"Error creating tables: {e}")


if __name__ == "__main__":
    create_db_and_tables()



import time
import uuid
from typing import Optional

from langchain_community.llms.llamafile import Llamafile
from openai.types.model import Model
from sqlalchemy import func
from sqlmodel import Field, Session, SQLModel, select

from timestep.config import Settings
from timestep.database import (  # ModelAliasSQLModel,; ModelSQLModel,; create_db_and_tables,
    AgentSQLModel,
    engine,
)

settings = Settings()

app_dir = settings.app_dir
default_tools = []

# class AgentService(object):
#     # models: dict[uuid.UUID] = {}
#     _shared_instance_state = {
#         "models": {},
#     }

#     def __new__(cls, *args, **kwargs):
#         obj = super(AgentService, cls).__new__(cls, *args, **kwargs)
#         obj.__dict__ = cls._shared_instance_state

#         return obj


# class ModelInstanceStoreSingleton(object):
#     _shared_model_instances: dict[str] = {}

#     def __new__(cls, *args, **kwargs):
#         obj = super(ModelInstanceStoreSingleton, cls).__new__(cls, *args, **kwargs)
#         obj.__dict__ = cls._shared_model_instances

#         # create_db_and_tables()

#         # try:
#         #     obj.create_model(
#         #         model_aliases=[
#         #             "gpt-3.5-turbo",
#         #             "gpt-3.5-turbo-1106",
#         #             "gpt-4-1106-preview",
#         #             "gpt-4-turbo",
#         #             "gpt-4o",
#         #             "gpt-4o-mini",
#         #             "llamafile",
#         #             "LLaMA_CPP",
#         #         ],
#         #     )

#         # except Exception as e:
#         #     print(f"Error creating model: {e}")

#         return obj

#     def create_model(self, model_aliases=[]):
#         with Session(engine) as session:
#             model = ModelSQLModel()

#             session.add(model)

#             for model_alias in model_aliases:
#                 model_alias = ModelAliasSQLModel(alias=model_alias, model_id=model.id)

#                 session.add(model_alias)

#             session.commit()
#             session.refresh(model)

#         model_id = str(model.id)
#         model_instance = Llamafile(
#             base_url=f"http://{settings.default_llamafile_host}:{settings.default_llamafile_port}",
#         )

#         for model_id in [model_id] + model_aliases:
#             self._shared_model_instances[model_id] = model_instance

#         return model

#     def delete_model(self, model_id):
#         # with Session(engine) as session:
#         #     model = session.get(ModelSQLModel, uuid.UUID(model_id))

#         #     session.delete(model)
#         #     session.commit()

#         # for model_id in [model_id] + model.aliases:
#         #     self._shared_model_instances.pop(model_id)

#         return model_id

#     def get_model(self, model_id_or_alias):
#         model_id = None

#         # try:
#         #     model_id = uuid.UUID(model_id_or_alias)

#         # except ValueError:
#         #     model_alias = model_id_or_alias

#         # with Session(engine) as session:
#         #     if model_id:
#         #         model = session.get(ModelSQLModel, model_id)

#         #     else:
#         #         model_alias = session.get(ModelAliasSQLModel, model_alias)

#         #         if model_alias:
#         #             model = session.get(ModelSQLModel, model_alias.model_id)

#         #         else:
#         #             raise ValueError(f"Model not found: {model_id_or_alias}")


#         return model

#     def get_model_instance(self, model_id):
#         return self._shared_model_instances.get(model_id)

#     def get_models(self):
#         with Session(engine) as session:
#             # models = session.exec(select(ModelSQLModel)).all()
#             # SELECT DISTINCT(model) FROM agents ORDER BY model DESC;
#             models = session.exec(select(AgentSQLModel.model)).distinct().order_by(AgentSQLModel.model.desc()).all()

#         return models


# model_instance_store = ModelInstanceStoreSingleton()
# model_instance_store.created_at = time.time()

# model_instance_store.create_model(
#     model_aliases=[
#         "gpt-3.5-turbo",
#         "gpt-3.5-turbo-1106",
#         "gpt-4-1106-preview",
#         "gpt-4-turbo",
#         "gpt-4o",
#         "gpt-4o-mini",
#         "llamafile",
#         "LLaMA_CPP",
#     ],
# )


async def delete_agent(id):
    with Session(engine) as session:
        agent = session.get(AgentSQLModel, uuid.UUID(id))

        session.delete(agent)
        session.commit()

    return agent


# async def delete_model(model_id):
#     try:
#         model_instance_store.delete_model(model_id)

#     except ValueError as e:
#         print(f"Error deleting model: {e}")

#     return model_id


async def get_agent(id: str | None = None, model: str | None = None):
    with Session(engine) as session:
        # agent = session.get(AgentSQLModel, uuid.UUID(id))
        if id:
            agent = session.exec(
                select(AgentSQLModel).where(AgentSQLModel.id == uuid.UUID(id))
            ).first()

        elif model:
            # SELECT MIN(created_at) AS created_at FROM agents WHERE model = 'gpt-4-1106-preview';
            agent = session.exec(
                select(AgentSQLModel).where(AgentSQLModel.model == model)
            ).first()

    return agent


async def get_agents(
    token_info: dict,
    user: str,
    after: Optional[str] = None,
    before: Optional[str] = None,
    limit: int = 20,
    order: str = "desc",
):
    # with Session(engine) as session:
    #     agents = session.exec(select(AgentSQLModel)).all()

    with Session(engine) as session:
        after_created_at = (
            session.exec(
                select(AgentSQLModel).where(AgentSQLModel.id == uuid.UUID(after))
            )
            .first()
            .created_at
            if after
            else 0
        )

        before_created_at = (
            session.exec(
                select(AgentSQLModel).where(AgentSQLModel.id == uuid.UUID(before))
            )
            .first()
            .created_at
            if before
            else func.now()
        )

        statement = (
            select(AgentSQLModel)
            .where(
                AgentSQLModel.created_at.between(after_created_at, before_created_at),
            )
            .order_by(
                AgentSQLModel.created_at.desc()
                if order == "desc"
                else AgentSQLModel.created_at.asc()
            )
            .limit(limit)
        )

        results = session.exec(statement)
        agents = results.all()

    return agents


async def get_default_agent():
    raise NotImplementedError


async def insert_agent(body):
    description = body.get("description")
    instructions = body.get("instructions")
    # model_alias = body.get("model")
    model = body.get("model")
    name = body.get("name")
    tools = body.get("tools", default_tools)

    # model = await retrieve_model(model_alias)

    agent = AgentSQLModel(
        description=description,
        instructions=instructions,
        # models=[{"id": model_id} for model_id in [model.id] + model.aliases],
        # model_id=model.id,
        model=model,
        name=name,
        tools=tools,
    )

    with Session(engine) as session:
        session.add(agent)
        session.commit()
        session.refresh(agent)

    return agent


# async def list_models():
#     models = model_instance_store.get_models()

#     return models


async def update_agent(id, body):
    with Session(engine) as session:
        agent = session.get(AgentSQLModel, uuid.UUID(id))

        # if "model" in body:
        #     model = await retrieve_model(body.get("model"))
        #     agent.models = [{"id": model_id} for model_id in [model.id] + model.aliases]

        agent.description = body.get("description", agent.description)
        agent.instructions = body.get("instructions", agent.instructions)
        agent.model = body.get("model", agent.model)
        agent.name = body.get("name", agent.name)
        agent.tools = body.get("tools", agent.tools)

        session.add(agent)
        session.commit()
        session.refresh(agent)

    return agent


# async def retrieve_model(model_id):
#     model = model_instance_store.get_model(model_id)

#     return model



import inspect

import typer

from timestep.config import Settings
from timestep.server import main as timestep_serve

settings = Settings()


def get_help_message():
    is_readme_context = inspect.getmodule(inspect.stack()[1][0]) is None

    return """
Timestep AI CLI - free, local-first, open-source AI
""" + (
        """
**Development Setup**:

```console
$ python3 -m pip install --upgrade pip
$ python3 -m pip install --user pipx
$ python3 -m pipx ensurepath
$ pipx install poetry==1.8.3
$ cp .env.example .env
$ direnv allow # See https://direnv.net/#getting-started
$ make
$ timestep up --dev
```

**Library Setup**:

```console
$ python3 -m pip install --upgrade pip
$ python3 -m pip install --user pipx
$ python3 -m pipx ensurepath
$ pipx install timestep
$ timestep up
```
"""
        if is_readme_context
        else ""
    )


typer_app = typer.Typer(
    help=get_help_message(),
    no_args_is_help=True,
)


@typer_app.callback()
def main():
    f"""
    Timestep AI CLI
    """


@typer_app.command()
def up(
    dev: bool = False,
    host: str = "0.0.0.0",
    # llamafile_path=f"./models/{default_llamafile_filename}", # TODO: namespace under llamafile, include port, etc.
    port: int = 8000,
):
    """
    Start up the Timestep AI platform.
    """

    typer.echo(f"Starting up the Timestep AI platform at http://{host}:{port}...")

    timestep_serve(
        dev=dev,
        host=host,
        # llamafile_path=llamafile_path,
        port=port,
    )


if __name__ == "__main__":
    typer_app()



import time
import uuid
from typing import List, Optional

from openai.types.beta.thread import Thread
from openai.types.beta.threads.message import Message, MessageContent
from openai.types.beta.threads.text import Text
from openai.types.beta.threads.text_content_block import TextContentBlock
from openai.types.model import Model
from sqlalchemy import func
from sqlmodel import Field, Session, SQLModel, select

from timestep.config import Settings
from timestep.database import MessageSQLModel, ThreadSQLModel, engine


async def insert_thread(*args, **kwargs):
    thread = ThreadSQLModel()

    with Session(engine) as session:

        session.add(thread)
        session.commit()
        session.refresh(thread)

    return thread


async def insert_message(*args, **kwargs):
    print("=== insert_message ===")
    print("args: ", args)
    print("kwargs: ", kwargs)

    attachments = kwargs.get("attachments", [])

    body_content = kwargs.get("body", {}).get("content")
    # content: List[MessageContent] = []
    content: List[dict] = []

    if body_content:
        content.append(
            # TextContentBlock(
            dict(
                # text=Text(
                text=dict(
                    annotations=[],
                    value=(
                        body_content
                        if type(body_content) == str
                        else body_content[0].get("text")
                    ),
                ),
                type="text",
            )
        )

    message = MessageSQLModel(
        attachments=attachments,
        content=content,
        role=kwargs.get("body", {}).get("role"),
        status="incomplete",
        thread_id=uuid.UUID(kwargs.get("thread_id")),
    )

    with Session(engine) as session:

        session.add(message)
        session.commit()
        session.refresh(message)

    return message


# def create_message(thread_id, create_message_request):  # noqa: E501
async def create_thread_message(body, token_info, thread_id, user):
    """Create a message.

     # noqa: E501

    :param thread_id: The ID of the [thread](/docs/api-reference/threads) to create a message for.
    :type thread_id: str
    :param create_message_request:
    :type create_message_request: dict | bytes

    :rtype: Union[MessageObject, Tuple[MessageObject, int], Tuple[MessageObject, int, Dict[str, str]]
    """
    # if connexion.request.is_json:
    #     create_message_request = CreateMessageRequest.from_dict(connexion.request.get_json())  # noqa: E501

    print("=== create_thread_message ===")
    raise NotImplementedError


# def get_thread(thread_id):  # noqa: E501
async def get_thread(token_info, thread_id, user):
    """Retrieves a thread.

     # noqa: E501

    :param thread_id: The ID of the thread to retrieve.
    :type thread_id: str

    :rtype: Union[ThreadObject, Tuple[ThreadObject, int], Tuple[ThreadObject, int, Dict[str, str]]
    """
    print("=== get_thread ===")
    raise NotImplementedError


async def get_thread_messages(
    token_info: dict,
    thread_id: str,
    user: str,
    after: Optional[str] = None,
    before: Optional[str] = None,
    limit: int = 20,
    order: str = "desc",
):
    """Returns a list of messages for a given thread.

     # noqa: E501

    :param thread_id: The ID of the [thread](/docs/api-reference/threads) the messages belong to.
    :type thread_id: str
    :param limit: A limit on the number of objects to be returned. Limit can range between 1 and 100, and the default is 20.
    :type limit: int
    :param order: Sort order by the &#x60;created_at&#x60; timestamp of the objects. &#x60;asc&#x60; for ascending order and &#x60;desc&#x60; for descending order.
    :type order: str
    :param after: A cursor for use in pagination. &#x60;after&#x60; is an object ID that defines your place in the list. For instance, if you make a list request and receive 100 objects, ending with obj_foo, your subsequent call can include after&#x3D;obj_foo in order to fetch the next page of the list.
    :type after: str
    :param before: A cursor for use in pagination. &#x60;before&#x60; is an object ID that defines your place in the list. For instance, if you make a list request and receive 100 objects, ending with obj_foo, your subsequent call can include before&#x3D;obj_foo in order to fetch the previous page of the list.
    :type before: str
    :param run_id: Filter messages by the run ID that generated them.
    :type run_id: str

    :rtype: Union[ListMessagesResponse, Tuple[ListMessagesResponse, int], Tuple[ListMessagesResponse, int, Dict[str, str]]
    """
    messages: List[Message] = []

    with Session(engine) as session:
        after_created_at = (
            session.exec(
                select(MessageSQLModel).where(MessageSQLModel.id == uuid.UUID(after))
            )
            .first()
            .created_at
            if after
            else 0
        )

        before_created_at = (
            session.exec(
                select(MessageSQLModel).where(MessageSQLModel.id == uuid.UUID(before))
            )
            .first()
            .created_at
            if before
            else func.now()
        )

        statement = (
            select(MessageSQLModel)
            .where(
                MessageSQLModel.thread_id == uuid.UUID(thread_id),
                MessageSQLModel.created_at.between(after_created_at, before_created_at),
            )
            .order_by(
                MessageSQLModel.created_at.desc()
                if order == "desc"
                else MessageSQLModel.created_at.asc()
            )
            .limit(limit)
        )

        results = session.exec(statement)
        messages = results.all()

    return messages


async def mutate_thread(*args, **kwargs):
    raise NotImplementedError


async def query_thread(*args, **kwargs):
    raise NotImplementedError



from typing import List

from openai.types import Model, ModelDeleted

from timestep.config import Settings
from timestep.database import AgentSQLModel
from timestep.services import agent_service

settings = Settings()


# async def delete_model(*args, **kwargs):
async def delete_model(model: str, token_info: dict, user: str):
    """Delete a fine-tuned model. You must have the Owner role in your organization to delete a model.

    :param model: The model to delete
    :type model: str

    :rtype: Union[DeleteModelResponse, Tuple[DeleteModelResponse, int], Tuple[DeleteModelResponse, int, Dict[str, str]]
    """
    agents: List[AgentSQLModel] = [
        agent
        for agent in await agent_service.get_agents(
            limit=None, token_info=token_info, user=user
        )
        if agent.model == model
    ]

    for agent in agents:
        await agent_service.delete_agent(id=agent.id)

    return ModelDeleted(
        id=model,
        deleted=True,
        object="model",
    ).model_dump(mode="json")


async def list_models(token_info: dict, user: str):
    """Lists the currently available models, and provides basic information about each one such as the owner and availability.

    :rtype: Union[ListModelsResponse, Tuple[ListModelsResponse, int], Tuple[ListModelsResponse, int, Dict[str, str]]
    """
    # models = await agent_service.list_models()
    agents: List[AgentSQLModel] = await agent_service.get_agents(
        limit=None, token_info=token_info, user=user
    )
    # models: List[Model] = []
    models_by_id = {}

    for agent in agents:
        if agent.model not in models_by_id:
            models_by_id[agent.model] = Model(
                # id=str(model.id),
                id=str(agent.model),
                created=agent.created_at.timestamp(),
                object="model",
                # owned_by=model.owned_by,
                owned_by=settings.openai_org_id,
            )

        else:
            models_by_id[agent.model].created = min(
                models_by_id[agent.model].created, agent.created_at.timestamp()
            )

    models = list(models_by_id.values())

    # models: List[Model] = [
    #     Model(
    #         # id=str(model.id),
    #         id=str(agent.model),
    #         created=agent.created_at.timestamp(),
    #         object="model",
    #         # owned_by=model.owned_by,
    #         owned_by=settings.openai_org_id,
    #     )
    #     for agent in agents
    # ]

    return {
        "data": [model.model_dump(mode="json") for model in models],
        "object": "list",
    }


async def retrieve_model(model: str, token_info: dict, user: str):
    """Retrieves a model instance, providing basic information about the model such as the owner and permissioning.

     # noqa: E501

    :param model: The ID of the model to use for this request
    :type model: str

    :rtype: Union[Model, Tuple[Model, int], Tuple[Model, int, Dict[str, str]]
    """
    # model = await agent_service.retrieve_model(model_id=model)
    agent: AgentSQLModel = await agent_service.get_agent(model=model)

    model: Model = Model(
        # id=str(model.id),
        # id=str(agent.id),
        id=str(agent.model),
        # created=model.created_at.timestamp(),
        created=agent.created_at.timestamp(),
        object="model",
        # owned_by=model.owned_by,
        owned_by=settings.openai_org_id,
    )

    return model.model_dump(mode="json")
